﻿using System;
using System.IO;

namespace TBL.CSharp.SaveData.LSD
{
    /// <summary>
    /// <see cref="LSD"/> 基础支持
    /// </summary>
    public abstract partial class Base
    {
        /// <summary>
        /// PNG Chunk 标识长度
        /// </summary>
        public const int ChunkSignSize = 4;
        
        /// <summary>
        /// 产品名称长度
        /// </summary>
        public const int ProductNameSize = 8;

        /// <summary>
        /// 版本信息长度
        /// </summary>
        public const int VersionInfoSize = 8;
        
        /// <summary>
        /// 产品名称
        /// </summary>
        protected readonly byte[] ProductName;

        /// <summary>
        /// CRC 校验表
        /// </summary>
        private static readonly uint[] CrcTable;

        /// <summary>
        /// 初始化 CRC 校验表
        /// </summary>
        static Base()
        {
            CrcTable = new uint[256];
            for(uint n = 0; n < CrcTable.Length; n++){
                var c = n;
                for(var k = 0; k < 8; k++){
                    if((c & 1) == 1)
                        c = 0xEDB88320 ^ ((c >> 1) & 0x7FFFFFFF);
                    else
                        c = ((c >> 1) & 0x7FFFFFFF);
                }
                CrcTable[n] = c;
            }
        }

        /// <summary>
        /// 构造
        /// </summary>
        /// <param name="productName">产品名称；必须是 8 字节的字符串</param>
        public Base(string productName)
        {
            var encoding = System.Text.Encoding.UTF8;
            // 检查产品名称长度
            if (encoding.GetByteCount(productName) != ProductNameSize)
            {
                throw new ArgumentOutOfRangeException(nameof(productName));
            }
            // 填充产品名称
            ProductName = encoding.GetBytes(productName);
        }

        /// <summary>
        /// 比较字节组
        /// </summary>
        public static bool BytesCompare(byte[] b1, byte[] b2, int count)
        {
            for (var i = 0; i < count; i++)
            {
                if (b1[i] != b2[i])
                    return false;
            }
            return true;
        }
        
        /// <summary>
        /// 比较字节组
        /// </summary>
        public static unsafe bool BytesCompare(byte* b1, byte[] b2, int count)
        {
            for (var i = 0; i < count; i++)
            {
                if (b1[i] != b2[i])
                    return false;
            }
            return true;
        }

        /// <summary>
        /// 读取流数据计算 CRC 校验值
        /// </summary>
        public static uint Crc32(Stream stream, int count, uint crc = 0)
        {
            var c = crc ^ 0xffffffff;
            for(var i = 0; i < count; i++){
                c = CrcTable[(c ^ stream.ReadByte()) & 255] ^ ((c >> 8) & 0xFFFFFF);
            }
            return c ^ 0xffffffff;
        }
    }
}