﻿using System;
using System.IO;

namespace TBL.CSharp.SaveData.LSD
{
    /// <summary>
    /// PNG 数据段
    /// </summary>
    public unsafe struct Chunk
    {
        /// <summary>
        /// 标识
        /// </summary>
        public fixed byte Sign[TBL.CSharp.SaveData.LSD.Base.ChunkSignSize];
        
        /// <summary>
        /// 数据区所属的流位置
        /// </summary>
        public readonly long DataPosition;
        
        /// <summary>
        /// 数据区的尺寸
        /// </summary>
        public readonly uint DataSize;

        /// <summary>
        /// Crc 校验码
        /// </summary>
        public readonly uint Crc;

        /// <summary>
        /// 段类型所属的流位置
        /// </summary>
        public long SignPosition => DataPosition - TBL.CSharp.SaveData.LSD.Base.ChunkSignSize;
        
        /// <summary>
        /// 段头部所属的流位置
        /// </summary>
        public long HeadPosition => SignPosition - sizeof(UInt32);

        /// <summary>
        /// 构造
        /// </summary>
        public Chunk(uint dataSize, byte[] sign, long dataPosition, uint crc)
        {
            DataSize = dataSize;
            for (var i = 0; i < TBL.CSharp.SaveData.LSD.Base.ChunkSignSize; i++)
            {
                Sign[i] = sign[i];
            }
            DataPosition = dataPosition;
            Crc = crc;
        }

        /// <summary>
        /// 检查数据段类型
        /// </summary>
        public readonly bool CheckChunkType(byte[] type)
        {
            fixed (byte * ptr = Sign)
            {
                return TBL.CSharp.SaveData.LSD.Base.BytesCompare(ptr, type, TBL.CSharp.SaveData.LSD.Base.ChunkSignSize);
            }
        }

        /// <summary>
        /// 检查 Crc 是否正确
        /// <para>流位置会被变更到此数据块</para>
        /// </summary>
        public bool CheckCrcCorrect(Stream stream)
        {
            stream.Seek(SignPosition, SeekOrigin.Begin);
            return TBL.CSharp.SaveData.LSD.Base.Crc32(stream, (int)(TBL.CSharp.SaveData.LSD.Base.ChunkSignSize + DataSize)) == Crc;
        }

        /// <summary>
        /// 尝试以 <see cref="LSD"/> 元数据块解析
        /// <para>流位置会被变更到此数据块</para>
        /// </summary>
        public bool TryParseAsLsdChunk(Stream stream, 
            out string productName,
            out string versionInfo,
            out bool isLittleEndian
            )
        {
            stream.Seek(DataPosition, SeekOrigin.Begin);
            if (!CheckChunkType(TBL.CSharp.SaveData.LSD.Base.Lsd.Signature))
            {
                productName = null;
                versionInfo = null;
                isLittleEndian = false;
                return false;
            }
            var encoding = System.Text.Encoding.UTF8;
            var buffer = new byte[Math.Max(TBL.CSharp.SaveData.LSD.Base.ProductNameSize, TBL.CSharp.SaveData.LSD.Base.VersionInfoSize)];
            stream.Read(buffer, 0, TBL.CSharp.SaveData.LSD.Base.ProductNameSize);
            productName = encoding.GetString(buffer, 0, TBL.CSharp.SaveData.LSD.Base.ProductNameSize);
            stream.Read(buffer, 0, TBL.CSharp.SaveData.LSD.Base.VersionInfoSize);
            versionInfo = encoding.GetString(buffer, 0, TBL.CSharp.SaveData.LSD.Base.VersionInfoSize);
            isLittleEndian = stream.ReadByte() == TBL.CSharp.SaveData.LSD.Base.Lsd.DefaultEndian.LittleEndian;
            stream.Seek(3, SeekOrigin.Current);
            stream.Seek(4, SeekOrigin.Current);
            stream.Seek(8, SeekOrigin.Current);
            return true;
        }
    }
}