﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

namespace TBL.CSharp.SaveData.LSD
{
    /// <summary>
    /// <see cref="LSD"/> 读取器
    /// </summary>
    public class Reader : TBL.CSharp.SaveData.LSD.Base
    {
        /// <summary>
        /// 缓冲区
        /// </summary>
        private readonly byte[] _buffer;
        
        /// <summary>
        /// 构造
        /// </summary>
        /// <param name="productName">产品名称；必须是 8 字节的字符串</param>
        public Reader(string productName) : base(productName)
        {
            _buffer = new byte[16];
        }

        /// <summary>
        /// 自动处理字节序差异从流中读取一个 <see cref="UInt32"/>
        /// </summary>
        public unsafe UInt32 ReadUInt32(BinaryReader reader)
        {
            if (BitConverter.IsLittleEndian)
            {
                var size = sizeof(UInt32);
                var buffer = stackalloc byte[size];
                for (var i = 0; i < size; i++)
                    buffer[size - i - 1] = reader.ReadByte();
                return *(UInt32*) buffer;
            }
            else
            {
                return reader.ReadUInt32();
            }
        }

        /// <summary>
        /// 从流中扫描所有数据段
        /// <para>扫描后，流的位置会变化，但不会关闭流</para>
        /// </summary>
        /// <param name="stream">要扫描的流</param>
        /// <param name="crcCheck">是否进行 Crc 检查</param>
        /// <returns>所有数据段信息</returns>
        /// <exception cref="BadImageFormatException">识别 PNG 文件头错误的异常</exception>
        /// <exception cref="InvalidDataException">未通过 <paramref name="crcCheck"/> 的异常</exception>
        public IEnumerable<Chunk> ScanAllChunks(Stream stream, bool crcCheck = true)
        {
            using (var reader = new BinaryReader(stream, Encoding.UTF8, true))
            {
                // 识别 PNG 文件头
                reader.Read(_buffer, 0, PngHeader.Signature.Length);
                if (!BytesCompare(_buffer, PngHeader.Signature, PngHeader.Signature.Length))
                    throw new BadImageFormatException();
                // 逐个读取数据段
                var isNotEnding = true;
                while(isNotEnding)
                {
                    // 段长度
                    var chunkSize = ReadUInt32(reader);
                    // 段类型
                    reader.Read(_buffer, 0, ChunkSignSize);
                    // 标记当前位置
                    var positionData = stream.Position;
                    // 前进到正确的位置
                    stream.Position = positionData + chunkSize;
                    // 段校验
                    var chunkCrc = ReadUInt32(reader);
                    // 尾部标记判断
                    if (BytesCompare(_buffer, PngEnd.Signature, ChunkSignSize))
                        isNotEnding = false;
                    // 记录位置、产出数据段信息
                    var nextChunkPosition = stream.Position;
                    var chunk = new Chunk(chunkSize, _buffer, positionData, chunkCrc);
                    // 未通过 Crc 检定
                    if (crcCheck && !chunk.CheckCrcCorrect(stream))
                        throw new InvalidDataException(nameof(crcCheck));
                    // 呼出当前位置的数据段信息
                    yield return chunk;
                    // 还原流位置以便继续解析
                    stream.Position = nextChunkPosition;
                }
            }
        }
    }
}