﻿using System;
using System.IO;
using System.Text;

namespace TBL.CSharp.SaveData.LSD
{
    /// <summary>
    /// <see cref="LSD"/> 写入器
    /// </summary>
    public class Writer : TBL.CSharp.SaveData.LSD.Base
    {
        /// <summary>
        /// 构造
        /// </summary>
        /// <param name="productName">产品名称；必须是 8 字节的字符串</param>
        public Writer(string productName) : base(productName)
        {
        }

        /// <summary>
        /// 自动处理字节序差异向流中写入一个 <see cref="UInt32"/>
        /// </summary>
        public unsafe void WriteUInt32(BinaryWriter writer, UInt32 value)
        {
            if (BitConverter.IsLittleEndian)
            {
                var size = sizeof(UInt32);
                var buffer = stackalloc byte[size];
                *(UInt32*) buffer = value;
                for (var i = 0; i < size; i++)
                    writer.Write(buffer[size - i - 1]);
            }
            else
            {
                writer.Write(value);
            }
        }

        /// <summary>
        /// 追加文件头
        /// </summary>
        public void AppendHeader(Stream target)
        {
            target.Write(PngHeader.Signature, 0, PngHeader.Signature.Length);
        }

        /// <summary>
        /// 追加一个数据段
        /// </summary>
        public void AppendChunk(byte[] sign, Stream target, Action<BinaryWriter> dataSource)
        {
            if (sign.Length != ChunkSignSize)
                throw new ArgumentException(nameof(sign));
            using (var writer = new BinaryWriter(target, Encoding.UTF8, true))
            {
                // 长度数据占位
                var positionSize = target.Position;
                WriteUInt32(writer, 0);
                // 写入标记数据
                var positionSign = target.Position;
                writer.Write(sign, 0, ChunkSignSize);
                // 写入实际数据
                var positionData = target.Position;
                dataSource?.Invoke(writer);
                // Crc 数据占位
                var positionCrc = target.Position;
                var dataSize = Convert.ToUInt32(positionCrc - positionData);
                // 写入长度数据
                target.Position = positionSize;
                WriteUInt32(writer, dataSize);
                // 写入 CRC
                {
                    target.Position = positionSign;
                    var crc = Crc32(target, (int)(ChunkSignSize + dataSize));
                    target.Position = positionCrc;
                    WriteUInt32(writer, crc);
                }
            }
        }

        /// <summary>
        /// 追加 <see cref="LSD"/> 元数据块
        /// </summary>
        public void AppendLsdChunk(Stream target, byte[] versionInfo, byte endian)
        {
            AppendChunk(Lsd.Signature, target, data =>
            {
                data.Write(ProductName);
                data.Write(versionInfo, 0, VersionInfoSize);
                data.Write(endian);
                for (var i = 0; i < 3; i++)
                    data.Write((byte)0);
                for (var i = 0; i < 4; i++)
                    data.Write((byte)0);
                for (var i = 0; i < 8; i++)
                    data.Write((byte)0);
            });
        }

        /// <summary>
        /// 追加尾数据段
        /// </summary>
        public void AppendEndChunk(Stream target) => AppendChunk(PngEnd.Signature, target, null);
    }
}